from . import converters
from . import enums
from . import exceptions


class NamedList(list):
    def __init__(self, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __getattribute__(self, item):
        # TODO: these names aren't amazingly clear
        item_mapping = {"event": 0, "when": 1, "match": 2, "match_events": 3, "period_when": 4, "period_when_time": 5,
                        "identifier": 6, "ignore_events": 7, "ignore_when": 8, "ignore_when_time": 9,
                        "match_period": 10, "event_period": 11}

        if item in item_mapping:
            return self[item_mapping[item]]

        return super().__getattribute__(item)


def construct_rule(event, when, match, other_event, period_when, period, track, exclude_event, exclude_period_when,
                   exclude_period, limit_period, review_period):
    if when not in enums.When:
        raise exceptions.InvalidWhen(f"{when} is not a valid value for when")

    if exclude_period_when not in enums.PeriodWhen:
        raise exceptions.InvalidPeriodWhen(f"{exclude_period_when} is not a valid value for when")

    if match not in enums.Match:
        raise exceptions.InvalidMatch

    try:
        period_in_seconds = converters.convert_to_seconds(period)
    except Exception:
        raise exceptions.InvalidPeriod

    try:
        exclude_period_in_seconds = converters.convert_to_seconds(exclude_period)
        if period_when == enums.PeriodWhen.BEFORE:
            exclude_period_in_seconds *= -1
    except Exception:
        raise exceptions.InvalidPeriod

    try:
        limit_period_in_seconds = converters.convert_to_seconds(limit_period)
    except Exception:
        raise exceptions.InvalidPeriod

    try:
        review_period_in_seconds = converters.convert_to_seconds(review_period)
    except Exception:
        raise exceptions.InvalidPeriod

    return [event, when, match, other_event, period_when, period_in_seconds, track, exclude_event,
            exclude_period_when, exclude_period_in_seconds, limit_period_in_seconds, review_period_in_seconds]


class RulesRegistry(object):
    def __init__(self):
        self.__rules = {}

    def register(self, rule_name, rule=None, **kwargs):
        if rule_name in self.__rules:
            raise exceptions.AlreadyRegistered()

        if not rule and kwargs:
            rule = construct_rule(**kwargs)
        elif rule:
            rule = construct_rule(*rule)

        self.__rules[rule_name] = NamedList(rule_name, rule)

    def unregister(self, rule_name):
        if rule_name not in self.__rules:
            raise exceptions.RuleNotRegistered

        del self.__rules[rule_name]

    def rules_for_event(self, event_type):
        rules = []
        for label, rule in self.__rules.items():
            if event_type.lower() == rule[0].lower():
                rules.append(rule)

        return rules

    @property
    def rules(self):
        return self.__rules


registry = RulesRegistry()




from datetime import timedelta

from django.db.models import Q

from . import BaseRiskMixin


class DjangoRiskMixin(BaseRiskMixin):
    @classmethod
    def load_related_events(cls, ruleset, match):
        """ Load more rules based on the criteria in the matching event """
        dates_q = Q()
        event_types_q = Q()
        filters = match.get_event_instance_filter()

        for rule in ruleset:
            if rule[9] >= 0:
                rule_start_time = match.risk_event_date + timedelta(seconds=rule[9])
                rule_end_time = match.risk_event_date + timedelta(seconds=rule[10])
            else:
                rule_start_time = match.risk_event_date + timedelta(seconds=rule[10])
                rule_end_time = match.risk_event_date + timedelta(seconds=rule[9])

            dates_q |= Q(**{f'{cls.event_date_field}__range': [rule_start_time, rule_end_time]})
            event_types_q |= Q(**{cls.event_type_field: rule[3]})

        return cls.objects.filter(**filters).filter(dates_q).filter(event_types_q)

    @classmethod
    def load_events_for_analysis(cls, event_type, start_time, end_time, *args, **kwargs):
        """ Load an initial set of rules """
        filters = {cls.event_type_field: event_type, f"{cls.event_date_field}__range": (start_time, end_time)}
        return cls.objects.filter(**filters)


class InvalidWhen(Exception): pass


class InvalidPeriodWhen(Exception): pass


class InvalidMatch(Exception): pass


class InvalidPeriod(Exception): pass


class AlreadyRegistered(Exception): pass


class RuleNotRegistered(Exception): pass



import logging
import typing
from datetime import timedelta, datetime

from .. import snoop

logger = logging.getLogger("porlock.risk_event")


class BaseRiskMixin:
    event_type_field = None
    event_date_field = None
    event_actor_field = None
    event_instance_field = None

    @property
    def risk_event_type(self) -> str:
        return getattr(self, self.event_type_field, None)

    @property
    def risk_event_date(self) -> datetime:
        return getattr(self, self.event_date_field, None)

    @property
    def risk_event_instance(self) -> typing.Any:
        return getattr(self, self.event_instance_field, None)

    def get_event_instance_filter(self) -> dict:
        return {self.event_instance_field: getattr(self, self.event_instance_field)}

    def construct_date_range(self, rule):
        if rule.ignore_when_time >= 0:
            rule_start_time = self.risk_event_date + timedelta(seconds=rule.ignore_when_time)
            rule_end_time = self.risk_event_date + timedelta(seconds=rule.match_period)
        else:
            rule_start_time = self.risk_event_date + timedelta(seconds=rule.match_period)
            rule_end_time = self.risk_event_date + timedelta(seconds=rule.ignore_when_time)

        return rule_start_time, rule_end_time

    @classmethod
    def load_events_for_analysis(cls, event_type, start_time, end_time, *args, **kwargs) -> typing.Iterable:
        """ Load events for analysis """
        raise NotImplementedError

    @classmethod
    def load_related_events(cls, ruleset, match) -> typing.Iterable:
        """ Load more rules based on the criteria in the matching event """
        raise NotImplementedError

    @classmethod
    def identify_risk(cls, events, aggregate_events=False):
        aggregated_events = {}

        for ruleset, match in snoop.find_rule_match(events):
            related_events = cls.load_related_events(ruleset, match)
            for risk_event in snoop.inspect_related_events(ruleset, match, related_events):
                if not aggregate_events:
                    logger.info("Risk Identified: [%s]", risk_event)
                    yield risk_event
                else:
                    if match not in aggregated_events:
                        aggregated_events[match] = []
                    aggregated_events[match].append(risk_event[2])

        if aggregate_events:
            for match, events in aggregated_events.items():
                yield "Aggregated Events", match, events # [0]



from datetime import timedelta

UNITS = {"s": "seconds", "m": "minutes", "h": "hours", "d": "days", "w": "weeks"}


def convert_to_seconds(s):
    count = int(s[:-1])
    unit = UNITS[s[-1]]
    td = timedelta(**{unit: count})
    return td.seconds + 60 * 60 * 24 * td.days




import typing

from .enums import Match
from .rules import registry


def find_rule_match(events) -> (str, typing.Any):
    rules = {}

    for event in events:
        if event.risk_event_type not in rules:
            rules[event.risk_event_type] = registry.rules_for_event(event.risk_event_type)

        if rules[event.risk_event_type]:
            yield rules[event.risk_event_type], event


def inspect_related_events(ruleset, match, events) -> (str, typing.Any, typing.Any):
    for rule in ruleset:
        rule_events = set([x.lower() for x in rule.match_events])
        rule_start_time, rule_end_time = match.construct_date_range(rule)

        if rule.match in [Match.ANY.value, Match.NONE.value]:
            for event in events:
                if rule.match == Match.ANY.value:
                    if event.risk_event_type.lower() in rule_events:
                        if rule_start_time <= event.risk_event_date <= rule_end_time:
                            yield rule, match, event
                elif rule.match == Match.NONE.value:
                    if event.risk_event_type.lower() not in rule_events:
                        if rule_start_time <= event.risk_event_date <= rule_end_time:
                            yield rule, match, event
        elif rule.match == Match.ALL.value:
            """ Match all event types """
            matched_events = []
            for event in events:
                if event.risk_event_type.lower() in rule_events:
                    if rule_start_time <= event.risk_event_date <= rule_end_time:
                        matched_events.append(event)

            matched_rule_events = set([event.risk_event_type.lower() for event in matched_events])
            if not rule_events.difference(matched_rule_events):
                for event in matched_events:
                    yield rule, match, event
        elif rule.match == Match.ONE.value:
            """ Match exactly one event type (there could be multiple events that match the event type) """
            matched_events = []
            for event in events:
                if event.risk_event_type.lower() in rule_events:
                    if rule_start_time <= event.risk_event_date <= rule_end_time:
                        matched_events.append(event)

            matched_rule_events = set([event.risk_event_type.lower() for event in matched_events])
            if len(matched_rule_events) == 1:
                for event in matched_events:
                    yield rule, match, event



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['porlock', 'porlock.mixins']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'porlock',
    'version': '0.0.1',
    'description': 'Find unwanted visitors by analyzing behavior',
    'long_description': 'porlock\n=======\nMonitor event logs for potentially fraudulent activity\n\nPremise\n--------------\nOftentimes, specific events in and of themselves don\'t indicate a\nrisk to your systems.  However, when viewed in the context of surrounding\nevents, we can identify potentially risky behavior.\n\nTake the following scenario, represented by the events below: \n- A user calls customer service and requests and one time token for login.\n- The user logs in, and the system requires a password update\n- A few days later, the user comes to the site (from a new IP) and requests a password reset\n- The user proceeds to change their password\n\n```\nEvent: otp_login, date: 2022-11-01 13:43:17 EST, user_id: 1, ip: 10.0.0.1\nEvent: password_change, date: 2022-11-01 13:45:53 EST, user_id: 1, ip: 10.0.0.1\nEvent: password_reset, date: 2022-11-07 17:31:22 EST, user_id: 1, ip: 192.168.0.1\nEvent: password_change, date: 2022-11-07 17:33:11 EST, user_id: 1, ip: 192.168.0.1\n```\n\nNone of the above events, in isolation, look particularly risky, and while it\nis possible that the user simply forgot their new password, when\nviewed along with other events, this sequence of events points to a potential account hijack.\n\nporlock provides a rule framework to help make finding these events and flagging them\npossible.  By identifying events, and subsequent matching events, porlock parses\nlogs over a specified time period to flag risky behavior.\n\nBasic Rules\n--------------\n\n- Rules are written in a simple list format \n  - Note: future versions may support yaml rules\n\n```\n\n    [\n       "Password Change After OTP Login",\n       "otp login",\n       "followed by",\n       "any",\n       ["password change"],\n       "after",\n       "2d",\n       "user",\n       ["password change"],\n       "before",\n       "1h",\n       "14d",\n       "30d"\n    ]\n```\n\n- `Rule Name`: A human readable description of the rule\n- `Event Name`: The name of the initial event to search for\n- `When`: `followed by` or `preceded by` (Currently only `followed by` is supported)\n- `Match`: What type of match to `any`, `all`, `none`, `one`\n- `Matching Events`: a list of events to match\n- `Period When`: whether or not this event should happen `before` or `after` the `Period` parameter\n- `Period`: How long `after` the initial event to start looking (or when to stop looking for `before`)\n- `Identifier`: The field to match events on.  This is typically either a user id, email, or IP address\n- `Secondary Event`: For future use\n- `Secondary Period When`: For future use\n- `Secondary Period`: For future use\n- `Match time frame`: The time frame after (for `followed by`) the initial event occurred to check for matching events\n- `Log time frame`: For future use; The time frame to return all events for a particular `identifier` from the logs\n\n## Parameters\n### When\n- `followed by` - the matching events should come after the initial event\n- `preceded by` - the matching events should come before the initial event (Note: This is still a work in progress)\n\n### Match\n- `any` - can match any of the listed events\n- `all` - must match all of the listed events\n- `none` - must match none of the listed events\n- `one` - must match one and only one of the listed events\n\n### Period When\n- `before` - indicates to check all events prior to the `Period` parameter\n- `after`- indicates to check all events after to the `Period` parameter\n\n### Period\n- The period should be specified by a number followed by one of `s` (seconds), `m` (minutes), `h` (hours), `d` (days), or `w` (weeks)',
    'author': 'David Graves',
    'author_email': 'dgraves@thesummitgrp.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import enum


class MemberMatchEnum(enum.EnumMeta):
    def __contains__(cls, item):
        return item in [v.value for v in cls.__members__.values()]


class When(enum.Enum, metaclass=MemberMatchEnum):
    FOLLOWED_BY = "followed by"
    PRECEDED_BY = "preceded by"


class Match(enum.Enum, metaclass=MemberMatchEnum):
    ANY = "any"
    ALL = "all"
    NONE = "none"
    ONE = "one"


class PeriodWhen(enum.Enum, metaclass=MemberMatchEnum):
    BEFORE = "before"
    AFTER = "after"

